#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>

static char Device[256] = "/dev/dsp";

static int Stereo=1;
static int Bits  =16;
static int Rate  =44100;

static void PrintVersionInfo(void)
{
    printf("ecat - audio output tool v"VERSION" (C) 1992,2002 Bisqwit\n");
}

#include <argh.h>
#include "cat.h"

int main(int argc, const char *const *argv)
{
    int fd, Heelp = 0;
    FILE *input = NULL;
    
    argh_init();
    
    argh_add_long("stereo", '2'); argh_add_bool('2'); argh_add_desc('2', "Specifies stereo sound. Redundant.", NULL);
    argh_add_long("mono",   'm'); argh_add_bool('m'); argh_add_desc('m', "Specifies mono sound.", NULL);
    argh_add_long("byte",   'b'); argh_add_bool('b'); argh_add_desc('b', "8-bit mode. Normally 16-bit mode.", NULL);
    argh_add_long("word",   '6'); argh_add_bool('6'); argh_add_desc('6', "16-bit mode. Redundant.", NULL);
    argh_add_long("rate",   'r'); argh_add_int('r',18,999999); argh_add_desc('r', "Specifies output rate. 44100 is default.", "<num>");
    argh_add_long("device", 'd'); argh_add_string('d',1,1023); argh_add_desc('d', "Specify device.", "<file>");
    argh_add_long("help",   'h'); argh_add_bool('h'); argh_add_desc('h', "Help", NULL);
    argh_add_long("version",'V'); argh_add_bool('V'); argh_add_desc('V', "Version information", NULL);
    
    argh_start_parse(argc, argv);
    for(;;)
    {
    	int c = argh_get_param();
    	if(c == -1)break;
    	switch(c)
    	{
            case 'V':
                PrintVersionInfo();
                return 0;
            case 'h':
                Heelp = 1;
                break;
            case '2':
            	if(argh_get_bool())++Stereo;else Stereo=0;
                break;
            case 'b':
                Bits = argh_get_bool() ? 8 : 16;
                break;
            case '6':
                Bits = argh_get_bool() ? 16 : 8;
                break;
            case 'm':
                if(argh_get_bool())Stereo = 0;else ++Stereo;
                break;
            case 'r':
                Rate = argh_get_int();
                break;
            case 'd':
            	strncpy(Device, argh_get_string(), sizeof Device);
            	Device[sizeof(Device)-1] = 0;
                break;
            default:
            {
                const char *s = argh_get_string();
                if(strcmp(s, "-") != 0)
                {
                	if(input)
                	{
                		fprintf(stderr, "Attempt to specify multiple files. Not supported.\n");
                		return -1;
                	}
                	input = fopen(s, "rb");
                	if(!input)
                	{
                		perror(s);
                		return -1;
                	}
                }
            }
    	}
    }
    if(!input)
    	input = stdin;
    
    if(!argh_ok())return -1;
    
    if(Heelp)
    {
        PrintVersionInfo();
        printf(
            "\nOutputs stdin to the sound device.\n");
        printf(
            "Usage: ecat [<options> [...]] [<inputfile>]\n"
            "Options:\n");
        argh_list_options();
        return 0;
    }
    
    fd = open(Device, O_WRONLY);
    if(ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &Bits) < 0
    || ioctl(fd, SNDCTL_DSP_STEREO,     &Stereo) < 0
    || ioctl(fd, SNDCTL_DSP_SPEED,      &Rate) < 0)
    {
        exit(-1);
    }
    dup2(fileno(input), 0);
    dup2(fd, 1);
    if(fd != 1)close(fd);
    if(fileno(input) != 0)fclose(input);
    
    argh_done();
    
    cat(); /* doesn't return */
    
    return -1;
}
