#include <cstdio>
#include <cstring>
#include <cerrno>
#include <fcntl.h>
#include <unistd.h>

#include "mixerstream.hh"

MixerStream::MixerStream(int serverfd, int pipefd) : problem(false)
{
    int pip[2];
    
    char InFormat[32] = "";
    char OutFormat[32]= "";
    char Options[128];
    
    char *dest = pipefd ? OutFormat : InFormat;
    for(unsigned a=0; a<sizeof InFormat; )
    {
        if(read(serverfd, &dest[a], 1) <= 0)break;
        if(dest[a] == '\r')continue;
        if(dest[a] == '\n') { dest[a] = '\0'; break; }
        ++a;
    }
    dest[(sizeof InFormat)-1] = '\0';
    
    for(unsigned a=0; a<sizeof Options; )
    {
        if(read(serverfd, &Options[a], 1) <= 0)break;
        if(Options[a] == '\r')continue;
        if(Options[a] == '\n') { Options[a] = '\0'; break; }
        ++a;
    }
    Options[(sizeof Options)-1] = '\0';
            
    pipe(pip);
    pid = fork();
    if(!pid)
    {
    	if(pipefd == 0)
    	{
    		dup2(serverfd, 0);
    		dup2(pip[1], 1);
    	}
    	else if(pipefd == 1)
    	{
    		dup2(pip[0],  0);
    		dup2(serverfd, 1);
    	}
        dup2(serverfd, 2); /* Redirect stderr to socket */
        
        close(pip[0]);
        close(pip[1]); /* close unused fds */
        close(serverfd);
        
        /* Nonblocking off, Blocking on */
        fcntl(pipefd, F_SETFL, fcntl(pipefd, F_GETFL) &~ O_NONBLOCK);
        execlp("resample", /*pipefd?"mixmon":"mixwrite" */"resample",
            InFormat,
            OutFormat,
            "-H", /* Resample internal option */
            Options[0] ? Options : NULL,
            NULL);
    }
    
    /* Take what we want, close the other */
    /* close unused fd s*/
    if(pipefd == 0)
    {
    	fd = pip[0];
    	close(pip[1]);
    }
    else
    {
    	fd = pip[1];
    	close(pip[0]);
    }
    close(serverfd);

    if(pipefd==1)
    {
    	/* mixmon? */
    	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
    }
}

MixerStream::~MixerStream()
{
	if(problems())
		close(fd);
}

bool MixerInput::Evaluatable() const
{
	unsigned freespace = (sizeof Buf) - bufsiz;
	if(!freespace)return false;
	return true;
}

void MixerInput::Evaluate()
{
	unsigned freespace = (sizeof Buf) - bufsiz;
	if(!freespace)return;
    int c = read(fd, Buf+bufsiz, freespace);
    /* Reading from pipe does _not_ block until whole size read */
    if(c <= 0) { problem = true; return; }
    bufsiz += c;
}

/* Acknowledge elements (throw them out) */
void MixerInput::Eat(unsigned count)
{
    if(count > size())return;
    unsigned skipcount = count*4;        
    bufsiz -= skipcount;
    memmove(Buf, Buf + skipcount, bufsiz);
}

void MixerOutput::Send(const stent *s, unsigned n)
{
    unsigned oldsize = Buf.size();
    if(oldsize > 0x20000) /* 128 kB */
    {
    	fprintf(stderr, "mixeroutput: Wasting %u bytes of buffered data\n",
    	                oldsize-head);
    	// tllin kylmsti skipataan jono,
    	// jotta pstn takaisin mukaan.
    	oldsize = head;
    }
    Buf.resize(oldsize + n*4);
    memcpy(&Buf[0] + oldsize, s, n*4);
}

void MixerOutput::Evaluate()
{
	unsigned count = pending();
	if(!count)return;
    int c = write(fd, &Buf[head], count);
    if(c < 0 && errno != EAGAIN) { problem = true; return; }
    if(c <= 0)return;
    head += c;
    // fprintf(stderr, "mixeroutput: wrote %u, head=%u, Buf.size()=%u\n", c,head,Buf.size());
    if(head >= 4096)
    {
    	/* chop it */
        Buf.erase(Buf.begin(), Buf.begin() + head);
        head = 0;
    }
}

SockData::SockData(Waiter &w)
{
    They    = w.They;
    AddrLen = w.AddrLen;
}
