#include <stdio.h>      /* for stderr      */
#include <errno.h>      /* for EADDRINUSE  */
#include <string.h>     /* for memset      */
#include <fcntl.h>      /* for fcntl       */
#include <sys/socket.h> /* for socket()    */
#include <sys/stat.h>   /* for O_NONBLOCK  */
#include <unistd.h>
#include <netinet/in.h> /* for sockaddr_in */

#include "logfun.h"
#include "serverfun.h"

int MakeListenerSocket(unsigned port)
{
    struct sockaddr_in MyAddr;
    int Plug, dummy;
    unsigned retries = 0;
    
    if((Plug = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        return -1;

Retry:
    MyAddr.sin_port = htons(port);              /* network order */
    MyAddr.sin_family = AF_INET;                /* host byte order */
    MyAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* auto-fill with my IP, network long  */
    memset((void *)&(MyAddr.sin_zero), 0, 8);   /* zero the rest of the struct */

    dummy = 1;
    setsockopt(Plug, SOL_SOCKET, SO_REUSEADDR, &dummy, sizeof dummy);
    if(bind(Plug, (struct sockaddr *)&MyAddr, sizeof(struct sockaddr)) < 0)
    {
        DumpTime();
        fprintf(stderr, "Bind port %u failed (errno=%s).\n", port, strerror(errno));
           
        if(errno == EADDRINUSE)
        {
            DumpTime();
            fprintf(stderr, "Retrying in 5 seconds\n");
            sleep(5);
            ++retries;
            goto Retry;
        }
        
        close(Plug);
        return -1;
    }
    if(retries)
    {
        DumpTime();
        fprintf(stderr, "Succeed");
        if(retries > 1)fprintf(stderr, " after %u retries", retries);
        fprintf(stderr, "\n");
    }

    /* Plug will listen. Number of max waiting connections = 10 */
    if(listen(Plug, 10) < 0)
    {
        close(Plug);
        return -1;
    }
    if(!retries)
    {
        DumpTime();
        fprintf(stderr, "Listening to port %u\n", port);
    }
    return Plug;
}

unsigned SafeRead(void *d, unsigned bcount)
{
    char *dd = (char *)d;
    unsigned r = 0;
    while(bcount > 0)
    {
        int res = read(0, dd, bcount);
        if(res <= 0)break;
        bcount -= res;
        r += res;
        dd += res;
    }
    return r;
}

void SafeWrite(const void *d, unsigned bcount)
{
    write(1, d, bcount);
}
